@@ -6,6 +6,7 @@ package DNS::ZoneParse;
 use 5.006;
 use Storable 'dclone';
 use POSIX 'strftime';
+use File::Basename;
 use vars qw($VERSION);
 use strict;
 use Carp;
@@ -14,11 +15,16 @@ use Carp;
 # the backslash, here.
 my @ESCAPABLE_CHARACTERS = qw/ ; " \\\\ /;
 
-$VERSION = '0.99';
+my $rr_class             = qr/(?:IN|HS|CH)/i;
+my $rr_ttl               = qr/(?:\d+[wdhms]?)+/i;
+
+$VERSION = '1.00';
 my (
     %dns_id,  %dns_soa, %dns_ns,  %dns_a,     %dns_cname, %dns_mx, %dns_txt,
     %dns_ptr, %dns_a4,  %dns_srv, %dns_hinfo, %dns_rp,    %dns_loc,
-    %dns_last_name, %unparseable_line_callback, %last_parse_error_count,
+    %dns_generate,
+    %dns_last_name, %dns_last_origin, %dns_last_class, %dns_last_ttl,
+    %dns_found_origins, %unparseable_line_callback, %last_parse_error_count,
 );
 
 my %possibly_quoted = map { $_ => undef } qw/ os cpu text mbox /;
@@ -33,6 +39,7 @@ sub new {
     if ( ref $unparseable_callback eq 'CODE' ) {
         $unparseable_line_callback{$self} = $unparseable_callback;
     }
+
     $self->_initialize();
     $self->_load_file( $file, $origin ) if $file;
     return $self;
@@ -72,7 +79,12 @@ sub DESTROY {
     delete $dns_rp{$self};
     delete $dns_loc{$self};
     delete $dns_id{$self};
+    delete $dns_generate{$self};
     delete $dns_last_name{$self};
+    delete $dns_last_origin{$self};
+    delete $dns_last_ttl{$self};
+    delete $dns_last_class{$self};
+    delete $dns_found_origins{$self};
     delete $unparseable_line_callback{$self};
     delete $last_parse_error_count{$self};
 }
@@ -94,6 +106,7 @@ sub AUTOLOAD {
      : $method eq 'hinfo'    ? $dns_hinfo{$self}
      : $method eq 'rp'       ? $dns_rp{$self}
      : $method eq 'loc'      ? $dns_loc{$self}
+     : $method eq 'generate' ? $dns_generate{$self}
      : $method eq 'zonefile' ? $dns_id{$self}->{ZoneFile}
      : $method eq 'origin'   ? $dns_id{$self}->{Origin}
      :                         undef;
@@ -144,12 +157,19 @@ sub new_serial {
 sub output {
     my $self     = shift;
     my $zone_ttl = $dns_soa{$self}{ttl} ? "\$TTL $dns_soa{$self}{ttl}" : '';
-    my $output   = "";
-    $output .= <<ZONEHEADER;
+    my $output   = '';
+    $output .= <<ZONEHEADER1;
 ;
 ;  Database file $dns_id{$self}->{ZoneFile} for $dns_id{$self}->{Origin} zone.
 ;	Zone version: $dns_soa{$self}->{serial}
 ;
+ZONEHEADER1
+
+    if ( $dns_soa{$self}->{'ORIGIN'} ne $dns_soa{$self}->{'origin'} ) {
+        $output .= "\n\$ORIGIN $dns_soa{$self}->{'ORIGIN'}\n\n";
+    }
+
+    $output .= <<ZONEHEADER2;
 
 $zone_ttl
 $dns_soa{$self}->{origin}		$dns_soa{$self}->{ttl}	IN  SOA  $dns_soa{$self}->{primary} $dns_soa{$self}->{email} (
@@ -163,72 +183,130 @@ $dns_soa{$self}->{origin}		$dns_soa{$self}->{ttl}	IN  SOA  $dns_soa{$self}->{pri
 ; Zone NS Records
 ;
 
-ZONEHEADER
+ZONEHEADER2
+
+    my @origins_to_process = grep {
+        if ( $_ eq $dns_soa{$self}->{'ORIGIN'} ) {
+            0;
+        } else {
+            1;
+        }
+    } keys %{ $dns_found_origins{$self} };
+    unshift @origins_to_process, $dns_soa{$self}->{'ORIGIN'};
+
+    foreach my $process_this_origin ( @origins_to_process ) {
+        if ( $process_this_origin ne $dns_soa{$self}->{'ORIGIN'} ) {
+            $output .= "\n\;\n\; $process_this_origin records\n\;\n\n";
+            $output .= "\$ORIGIN $process_this_origin\n\n";
+        }
 
     foreach my $o ( @{ $dns_ns{$self} } ) {
         next unless defined $o;
+        next unless $o->{'ORIGIN'} eq $process_this_origin;
         $self->_escape_chars( $o );
         $output .= "$o->{name}	$o->{ttl}	$o->{class}	NS	$o->{host}\n";
     }
 
-    $output .= "\n\;\n\; Zone MX Records\n\;\n\n";
     foreach my $o ( @{ $dns_mx{$self} } ) {
         next unless defined $o;
+        next unless $o->{'ORIGIN'} eq $process_this_origin;
         $self->_escape_chars( $o );
-        $output .= "$o->{name}	$o->{ttl}	$o->{class}	MX	$o->{priority} " . " $o->{host}\n";
+        $output .= "$o->{name}	$o->{ttl}	$o->{class}	MX	$o->{priority} $o->{host}\n";
     }
 
-    $output .= "\n\;\n\; Zone Records\n\;\n\n";
     foreach my $o ( @{ $dns_a{$self} } ) {
         next unless defined $o;
+        next unless $o->{'ORIGIN'} eq $process_this_origin;
         $self->_escape_chars( $o );
         $output .= "$o->{name}	$o->{ttl}	$o->{class}	A	$o->{host}\n";
     }
     foreach my $o ( @{ $dns_cname{$self} } ) {
         next unless defined $o;
+        next unless $o->{'ORIGIN'} eq $process_this_origin;
         $self->_escape_chars( $o );
         $output .= "$o->{name}	$o->{ttl}	$o->{class}	CNAME	$o->{host}\n";
     }
     foreach my $o ( @{ $dns_a4{$self} } ) {
         next unless defined $o;
+        next unless $o->{'ORIGIN'} eq $process_this_origin;
         $self->_escape_chars( $o );
         $output .= "$o->{name}	$o->{ttl}	$o->{class}	AAAA	$o->{host}\n";
     }
     foreach my $o ( @{ $dns_txt{$self} } ) {
         next unless defined $o;
+        next unless $o->{'ORIGIN'} eq $process_this_origin;
         $self->_escape_chars( $o );
         $output .= qq[$o->{name}	$o->{ttl} $o->{class} TXT	"$o->{text}"\n];
     }
     foreach my $o ( @{ $dns_ptr{$self} } ) {
         next unless defined $o;
+        next unless $o->{'ORIGIN'} eq $process_this_origin;
         $self->_escape_chars( $o );
         $output .= "$o->{name}	$o->{ttl}	$o->{class}	PTR		$o->{host}\n";
     }
     foreach my $o ( @{ $dns_srv{$self} } ) {
         next unless defined $o;
+        next unless $o->{'ORIGIN'} eq $process_this_origin;
         $self->_escape_chars( $o );
         $output .= "$o->{name}	$o->{ttl}	$o->{class}	SRV	$o->{priority}	$o->{weight}	$o->{port}	$o->{host}\n";
     }
     foreach my $o ( @{ $dns_hinfo{$self} } ) {
         next unless defined $o;
+        next unless $o->{'ORIGIN'} eq $process_this_origin;
         $self->_escape_chars( $o );
         $output .= "$o->{name}	$o->{ttl}	$o->{class}	HINFO	$o->{cpu}   $o->{os}\n";
     }
     foreach my $o ( @{ $dns_rp{$self} } ) {
         next unless defined $o;
+        next unless $o->{'ORIGIN'} eq $process_this_origin;
         $self->_escape_chars( $o );
         $output .= "$o->{name}	$o->{ttl}	$o->{class}	RP	$o->{mbox}  $o->{text}\n";
     }
     foreach my $o ( @{ $dns_loc{$self} } ) {
         next unless defined $o;
+        next unless $o->{'ORIGIN'} eq $process_this_origin;
         $self->_escape_chars( $o );
         $output .= "$o->{name}	$o->{ttl}	$o->{class}	LOC	$o->{d1}	$o->{m1}	$o->{s1}	$o->{NorS}	";
         $output .= "$o->{d2}	$o->{m2}	$o->{s2}	$o->{EorW}	";
         $output .= "$o->{alt}	$o->{siz}	$o->{hp}	$o->{vp}\n";
     }
+    foreach my $o ( @{ $dns_generate{$self} } ) {
+        next unless defined $o;
+        next unless $o->{'ORIGIN'} eq $process_this_origin;
+        $self->_escape_chars( $o );
+        $output .= "\$GENERATE $o->{range}  $o->{lhs}  $o->{ttl}  $o->{class}  $o->{type}  $o->{rhs}\n";
+    }
+
+    }
+
     return $output;
 }
 
+sub fqname {
+    my ( $self, $record_ref ) = @_;
+
+    # Is this an SOA record?
+    if ( $record_ref->{'origin'} ) {
+        if ( ( $record_ref->{'origin'} eq '@' ) || ( $record_ref->{'origin'} =~ /\.$/ ) ) {
+            return $record_ref->{'ORIGIN'};
+        } else {
+            if ( $record_ref->{'ORIGIN'} =~ /^\./ ) {
+                return $record_ref->{'origin'} . $record_ref->{'ORIGIN'};
+            }
+            return $record_ref->{'origin'} . '.' . $record_ref->{'ORIGIN'};
+        }
+    } else {
+        if ( $record_ref->{'name'} eq '@' ) {
+            return $record_ref->{'ORIGIN'};
+        } else {
+            if ( $record_ref->{'ORIGIN'} =~ /^\./ ) {
+                return $record_ref->{'name'} . $record_ref->{'ORIGIN'};
+            }
+            return $record_ref->{'name'} . '.' . $record_ref->{'ORIGIN'};
+        }
+    }
+}
+
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
 # Private Methods
 # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
@@ -247,7 +325,13 @@ sub _initialize {
     $dns_srv{$self}       = [];
     $dns_hinfo{$self}     = [];
     $dns_rp{$self}        = [];
-    $dns_last_name{$self} = '@';
+    $dns_loc{$self}       = [];
+    $dns_generate{$self}  = [];
+    $dns_last_name{$self} = undef;
+    $dns_last_origin{$self} = undef;
+    $dns_last_ttl{$self} = undef;
+    $dns_last_class{$self} = 'IN'; # Class defaults to IN.
+    $dns_found_origins{$self} = {};
     $last_parse_error_count{$self} = 0;
     return 1;
 }
@@ -259,8 +343,9 @@ sub _load_file {
         $zone_contents = $$zonefile;
     } else {
         my $inZONE;
-        if ( open( $inZONE, $zonefile ) ) {
-            $zone_contents = do { local $/; <$inZONE> };
+        if ( open( $inZONE, '<', $zonefile ) ) {
+            local $/;
+            $zone_contents = <$inZONE>;
             close( $inZONE );
         } else {
             croak qq[DNS::ZoneParse Could not open input file: "$zonefile":$!];
@@ -275,12 +360,41 @@ sub _parse {
     my ( $self, $zonefile, $contents, $origin ) = @_;
     $self->_initialize();
 
-    my $chars = qr/[a-z\-\.0-9]+/i;
-    $contents =~ /Database file ($chars)( dns)? for ($chars) zone/sio;
-    $dns_id{$self} = $self->_massage( {
-            ZoneFile => $1 || $zonefile,
-            Origin   => $3 || $origin,
-    } );
+    # Here's how we auto-detect the zonefile and origin. Note, the zonefile is
+    # only used to print out a comment in the file, so its okay if we're
+    # inaccurate. First, prefer what the user configures. Next, try to read a
+    # comment we would have written if we wrote the file out in the past.
+    # Finally, pick up any SOA or $ORIGIN statements present in the file.
+    if ( ref( $zonefile ) eq 'SCALAR' ) { $zonefile = ''; }
+
+    if ( !$origin || !$zonefile ) {
+        # I don't know why the ( dns)? capture is there, perhaps at one point
+        # this module wrote a different header comment? I'll leave it as to
+        # preserve whatever backwards compatability this affords us...
+        $contents =~ /^\s*;\s*Database file (\S+)( dns)? for (\S+) zone/im;
+        if ( !$origin && $3 ) { $origin = $3; }
+        if ( !$zonefile && $1 ) { $zonefile = $1; }
+    }
+
+    if ( $zonefile ) {
+        $zonefile = basename( $zonefile );
+    } else {
+        $zonefile = 'unknown';
+    }
+
+    if ( $origin ) {
+        # A trite way of insuring there is a trailing dot on the origin. It's
+        # really important you supply a trailing . in an origin when you mean
+        # it.
+        $origin =~ s/([^.])$/$1./;
+    } else {
+        $origin = '';
+    }
+
+    $dns_id{$self} = {
+        ZoneFile => $zonefile,
+        Origin   => $origin,
+    };
 
     my $records = $self->_clean_records( $contents );
 
@@ -289,24 +403,31 @@ sub _parse {
      . join( '|', map { "\\\\$_" } @ESCAPABLE_CHARACTERS ) . ')';
 
     # The above, but adds the literal '.' character.
-    my $valid_name_char        = qr/(?:$valid_name_start_char|\.)/o;
+    my $valid_name_char        = qr/(?:$valid_name_start_char|[\.\\])/o;
+    my $valid_txt_char         = qr/\S+/o;
+    my $valid_quoted_txt_char  = qr/.+/o;
     # Like the above, but adds whitespace (space and tabs) too.
-    my $valid_quoted_name_char = qr/(?:$valid_name_start_char|[. ;\t()])/o;
-    my $valid_name             = qr/$valid_name_start_char(?:$valid_name_char|\.)*/o;
+    my $valid_quoted_name_char = qr/(?:$valid_name_start_char|[. ;\t()\\])/o;
+    my $valid_name             = qr/$valid_name_start_char$valid_name_char*/o;
     my $valid_ip6              = qr/[\@a-zA-Z_\-\.0-9\*:]+/;
-    my $rr_class               = qr/\b(?:IN|HS|CH)\b/i;
     my $rr_type                = qr/\b(?:NS|A|CNAME)\b/i;
-    my $rr_ttl                 = qr/(?:\d+[wdhms]?)+/i;
-    my $ttl_cls                = qr/(?:($rr_ttl)\s)?(?:($rr_class)\s)?/o;
-    my $last_name              = $dns_id{$self}->{Origin} || '@';
+    #my $ttl_cls                = qr/(?:($rr_ttl)\s)?(?:($rr_class)\s)?/o;
+    my $ttl_cls                = qr/(?:\b((?:$rr_ttl)|(?:$rr_class))\s)?(?:\b((?:$rr_class)|(?:$rr_ttl))\s)?/o;
+    my $generate_range         = qr{\d+\-\d+(?:/\d+)?};
+    my $last_good_line;
 
     foreach ( @$records ) {
-        TRACE( "parsing line <$_>" );
+        #TRACE( "parsing line <$_>" );
 
         # It's faster to skip blank lines here than to remove them inside
         # _clean_records.
         next if /^\s*$/;
 
+        # The below is inside of an eval block to catch possible errors
+        # found inside _massage and propagate them up properly.
+        eval {
+        local $SIG{__DIE__} = 'DEFAULT';
+
         if (
             /^($valid_name)? \s+         # host
               $ttl_cls                   # ttl & class
@@ -325,7 +446,7 @@ sub _parse {
                     name  => $name,
                     class => $class,
                     host  => $host,
-                    ttl   => $ttl
+                    ttl   => $ttl,
              } );
         } elsif (
             /^($valid_name)? \s+
@@ -341,7 +462,7 @@ sub _parse {
                     name  => $name,
                     class => $class,
                     host  => $host,
-                    ttl   => $ttl
+                    ttl   => $ttl,
              } );
         } elsif (
             /^($valid_name)? \s+
@@ -360,7 +481,7 @@ sub _parse {
                     priority => $pri,
                     host     => $host,
                     ttl      => $ttl,
-                    class    => $class
+                    class    => $class,
              } );
         } elsif (
             /^($valid_name)? \s+
@@ -383,37 +504,44 @@ sub _parse {
                     port     => $port,
                     host     => $host,
                     ttl      => $ttl,
-                    class    => $class
+                    class    => $class,
              } );
         } elsif (
             /^($valid_name) \s+
                  $ttl_cls
                  SOA \s+
                  ($valid_name) \s+
-                 ($valid_name) \s*    # Unsure if a space is required here.
-                 \(?\s*               # Or here.
-                     ($rr_ttl) \s+
-                     ($rr_ttl) \s+
-                     ($rr_ttl) \s+
-                     ($rr_ttl) \s+
-                     ($rr_ttl) \s*
-                 \)?
+                 ($valid_name) \s+
+                 ($rr_ttl) \s+
+                 ($rr_ttl) \s+
+                 ($rr_ttl) \s+
+                 ($rr_ttl) \s+
+                 ($rr_ttl)
                /ixo
          )
         {
             # SOA record
-            my $ttl = $dns_soa{$self}->{ttl} || $2 || '';
             $dns_soa{$self} = $self->_massage( {
                     origin     => $1,
-                    ttl        => $ttl,
+                    ttl        => $2,
+                    class      => $3,
                     primary    => $4,
                     email      => $5,
                     serial     => $6,
                     refresh    => $7,
                     retry      => $8,
                     expire     => $9,
-                    minimumTTL => $10
+                    minimumTTL => $10,
             } );
+
+            if ( !$origin ) {
+                $origin = $1;
+                $dns_id{$self} = {
+                    ZoneFile => $zonefile,
+                    Origin   => $origin,
+                };
+            }
+
         } elsif (
             /^($valid_name)? \s+
                 $ttl_cls
@@ -428,13 +556,13 @@ sub _parse {
                     name  => $1,
                     class => $3,
                     ttl   => $2,
-                    host  => $4
+                    host  => $4,
              } );
         } elsif (
             /($valid_name)? \s+
                 $ttl_cls
                 TXT \s+
-                ("$valid_quoted_name_char*(?<!\\)"|$valid_name_char+)
+                ("$valid_quoted_txt_char*(?<!\\)"|$valid_txt_char+)
             /ixo
         ) {
             push @{ $dns_txt{$self} },
@@ -442,20 +570,23 @@ sub _parse {
                     name  => $1,
                     ttl   => $2,
                     class => $3,
-                    text  => $4
+                    text  => $4,
              } );
         } elsif (
-            /\$TTL \s+
+            /^\s*\$TTL \s+
                 ($rr_ttl)
             /ixo
         ) {
-            $dns_soa{$self}->{ttl} = $1;
+            if ( !defined $dns_soa{$self} ) {
+                $dns_soa{$self}->{ttl} = $1;
+            }
+            $dns_last_ttl{$self} = $1;
         } elsif (
             /^($valid_name)? \s+
                  $ttl_cls
                  HINFO \s+
-                 ("$valid_quoted_name_char*(?<!\\)"|$valid_name_char+) \s+
-                 ("$valid_quoted_name_char*(?<!\\)"|$valid_name_char+)
+                 ("$valid_quoted_txt_char*(?<!\\)"|$valid_txt_char+) \s+
+                 ("$valid_quoted_txt_char*(?<!\\)"|$valid_txt_char+)
                /ixo
          )
         {
@@ -465,7 +596,7 @@ sub _parse {
                     ttl   => $2,
                     class => $3,
                     cpu   => $4,
-                    os    => $5
+                    os    => $5,
              } );
         } elsif (
             /^($valid_name)? \s+
@@ -482,7 +613,7 @@ sub _parse {
                     ttl   => $2,
                     class => $3,
                     mbox  => $4,
-                    text  => $5
+                    text  => $5,
              } );
         } elsif (
             /^($valid_name)? \s+
@@ -520,15 +651,57 @@ sub _parse {
                     siz   => $13,
                     hp    => $14,
                     vp    => $15,
-
              } );
+
+        } elsif ( /^\s*\$ORIGIN\s+($valid_name_char+)/io ) {
+            my $new_origin = $1;
+            # We could track each origins origin, all the way down, but what
+            # would that get us? Madness, surely.
+            if ( $new_origin !~ /\.$/ ) {
+                if ( $dns_last_origin{$self} =~ /^\./ ) {
+                    $new_origin .= $dns_last_origin{$self};
+                } else {
+                    $new_origin .= '.' . $dns_last_origin{$self};
+                }
+            }
+            $dns_last_origin{$self} = $new_origin;
+            $dns_found_origins{$self}->{ $new_origin } = 1;
+
+        } elsif ( /^ \s* \$GENERATE \s+
+                   ($generate_range) \s+     # range
+                   ($valid_name) \s+         # lhs
+                   (?:($rr_ttl) \s+)?        # ttl
+                   (?:($rr_class) \s+)?      # class
+                   ([a-z]+) \s+              # type
+                   ($valid_name)             # rhs
+                 /ixo
+         )
+        {
+            push @{ $dns_generate{$self} },
+             $self->_massage( {
+                    range  => $1,
+                    lhs    => $2,
+                    ttl    => $3,
+                    class  => $4,
+                    type   => $5,
+                    rhs    => $6,
+            } );
+
         } else {
+            die "Unknown record type\n";
+        }
+
+        }; # End of eval block.
+        if ( $@ ) {
+            chomp $@;
             $last_parse_error_count{$self}++;
             if ( $unparseable_line_callback{$self} ) {
-                $unparseable_line_callback{$self}->( $self, $_ );
+                $unparseable_line_callback{$self}->( $self, $_, $@, $last_good_line );
             } else {
-                carp "Unparseable line\n  $_\n";
+                carp "Unparseable line ($@)\n  $_\n";
             }
+        } else {
+            $last_good_line = $_;
         }
     }
     return 1;
@@ -546,9 +719,9 @@ sub _clean_records {
     my @lines;
 
     $zone =~ s/\r\n/\n/sg;
-    $zone =~ s/^\s*;.*$//mg;    # Remove lines with nothing but comments.
     $zone =~ s{[ \t]+}{ }g;     # Collapse whitespace, turn TABs to spaces.
 
+    # Trim comments, handle parentheses and some escape sequences.
     while (1) {
         my $c = substr( $zone, $x, 1 );
 
@@ -596,29 +769,136 @@ sub _clean_records {
 }
 
 sub _massage {
-    my $self   = shift;
-    my $record = shift;
+    my ( $self, $record ) = @_;
 
-    my $last_name = \$dns_last_name{$self};
+    foreach my $r ( keys %$record ) {
+        if ( !defined $record->{$r} ) {
+            $record->{$r} = '';
+            next;
+        }
+        if ( exists $possibly_quoted{$r} ) {
+            ( $record->{$r} =~ s/^"// ) && ( $record->{$r} =~ s/"$// );
+        }
 
-    foreach ( keys %$record ) {
-        $record->{$_} = '' unless defined $record->{$_};
-        $record->{$_} = uc $record->{$_} if $_ eq 'class';
-        $record->{$_} =~ s/\\(.)/$1/g;
-        if ( exists $possibly_quoted{$_} ) {
-            $record->{$_} =~ s/^"//;
-            $record->{$_} =~ s/"$//;
+        # We return email addresses just as they are in the file... for better
+        # or worse (mostly for backwards compatability reasons).
+        if ( $r ne 'email' && $r ne 'mbox' ) {
+            while ( $record->{$r} =~ m/\\/g ) {
+                my $pos = pos( $record->{$r} );
+                my $escape_char = substr( $record->{$r}, $pos, 1 );
+                if ( $escape_char =~ /\d/ ) {
+                    $escape_char = substr( $record->{$r}, $pos, 3 );
+                    # Max oct value that converts to 255 in dec.
+                    if ( ( $escape_char =~ /^\d{3}$/ ) && ( $escape_char <= 377 ) ) {
+                        substr( $record->{$r}, $pos - 1, 4 ) = chr( oct( $escape_char ) );
+                    } else {
+                        die "Invalid escape sequence\n";
+                    }
+                } else {
+                    # Not followed by a digit, so just remove the backslash.
+                    # Like BIND does...
+                    substr( $record->{$r}, $pos - 1, 2 ) = $escape_char;
+                }
+                pos( $record->{$r} ) = $pos;
+            }
         }
     }
 
-    return $record unless exists $record->{name};
-    if ( length $record->{name} ) {
-        $$last_name = $record->{name};
-    } else {
-        TRACE( "Record has no name, using last name" );
-        $record->{name} = $$last_name;
+    if (
+        ( ( $record->{'class'} =~ $rr_class ) && ( $record->{'ttl'} =~ $rr_class ) )
+        ||
+        ( ( $record->{'class'} =~ $rr_ttl   ) && ( $record->{'ttl'} =~ $rr_ttl   ) )
+    ) {
+        die "Invalid ttl/class values!\n";
+    };
+
+    if ( ( $record->{'class'} =~ $rr_ttl ) || ( $record->{'ttl'} =~ $rr_class ) ) {
+        my $x = $record->{'class'};
+        $record->{'class'} = $record->{'ttl'};
+        $record->{'ttl'} = $x;
     }
-    DUMP( "Record parsed", $record );
+
+    if ( $record->{'class'} ) {
+        $record->{'class'} = uc $record->{'class'};
+        $dns_last_class{$self} = $record->{'class'};
+    } else {
+        # This case should never happen, because we supply a default.
+        #if ( !defined $dns_last_class{$self} ) {
+        #    die "No class defined!\n";
+        #}
+        $record->{'class'} = $dns_last_class{$self};
+    }
+
+    if ( $record->{'ttl'} ) {
+        $record->{'ttl'} = uc $record->{'ttl'};
+        $dns_last_ttl{$self} = $record->{'ttl'};
+    } else {
+        if ( !defined $dns_last_ttl{$self} ) {
+            die "No ttl defined!\n";
+        }
+        $record->{'ttl'} = $dns_last_ttl{$self};
+    }
+
+    # This is silly, but we don't know what type of record we are massaging at
+    # this point. We can detect an SOA record because it's the only type that
+    # supplies this value, which is what we need to do here to properly set
+    # the owner.
+    if ( exists $record->{'minimumTTL'} ) {
+        $dns_last_name{$self} = $record->{'origin'};
+
+        if ( $record->{'origin'} eq '@' ) {
+            # We encountered a @ SOA line without an origin directive above
+            # it, so we will try and guess the origin.
+            if ( !$dns_last_origin{$self} ) {
+                if ( !$dns_id{$self}->{'Origin'} ) {
+                    die "Unknown origin\n";
+                }
+                $dns_last_origin{$self} = $dns_id{$self}->{'Origin'};
+            }
+            $record->{'ORIGIN'} = $dns_last_origin{$self};
+        } else {
+            my $new_origin = $record->{'origin'};
+
+            # Similar to above, it's origins all the way down. Don't bother
+            # tracking each separately, just collapse them all into the
+            # current origin.
+            if ( $new_origin =~ /\.$/ ) {
+                # If no one has set an $ORIGIN before, we need to use the SOA
+                # line to do it.
+                if ( !$dns_last_origin{$self} ) {
+                    $dns_last_origin{$self} = $new_origin;
+                }
+            }
+            # Now we have a valid ORIGIN for this SOA, so assign it.
+            $record->{'ORIGIN'} = $dns_last_origin{$self};
+        }
+        # Alright, make sure we know we found this origin.
+        $dns_found_origins{$self}->{ $record->{'ORIGIN'} } = 1;
+
+    # Not an SOA record.
+    } else {
+
+        # The silliness continues: only $GENERATE directives have a lhs, and
+        # don't need a 'name'.
+        if ( !exists $record->{'lhs'} ) {
+            if ( $record->{'name'} ) {
+                $dns_last_name{$self} = $record->{'name'};
+            } else {
+                #TRACE( "Record has no name, using last name" );
+                if ( !$dns_last_name{$self} ) {
+                    die "No current owner name\n";
+                }
+                $record->{'name'} = $dns_last_name{$self};
+            }
+        }
+
+        if ( !$dns_last_origin{$self} ) {
+            die "Unknown origin\n";
+        } else {
+            $record->{'ORIGIN'} = $dns_last_origin{$self};
+        }
+    }
+    #DUMP( "Record parsed", $record );
     return $record;
 }
 
@@ -678,12 +958,11 @@ nmake is available at http://download.microsoft.com/download/vc15/Patch/1.52/W95
 =head1 DESCRIPTION
 
 This module will parse a Zone File and put all the Resource Records (RRs)
-into an anonymous hash structure. At the moment, the following types of 
-RRs are supported: SOA, NS, MX, A, CNAME, TXT, PTR, HINFO, and RP. It could
-be useful for maintaining DNS zones, or for transferring DNS zones to other
-servers. If you want to generate an XML-friendly version of your zone files,
-it is easy to use XML::Simple with this module once you have parsed the
-zone file.
+into an anonymous hash structure. Various record types are supported, see the
+L<methods> section for details. It could be useful for maintaining DNS zones,
+or for transferring DNS zones to other servers. If you want to generate an
+XML-friendly version of your zone files, it is easy to use XML::Simple with
+this module once you have parsed the zone file.
 
 DNS::ZoneParse scans the DNS zone file - removes comments and seperates
 the file into its constituent records. It then parses each record and
@@ -726,13 +1005,14 @@ These methods return references to the resource records. For example:
 
 Returns the mx records in an array reference.
 
-All records have the following properties: 'ttl', 'class', 'host', 'name'.
+All records (except SOA) have the following properties: 'ttl', 'class',
+'host', 'name', 'ORIGIN'.
 
 MX records also have a 'priority' property.
 
 SRV records also have 'priority', 'weight' and 'port' properties.
 
-TXT records also have a 'text' property representing the record's "txt-data"
+TXT records also have a 'text' property representing the record's 'txt-data'
 descriptive text.
 
 HINFO records also have 'cpu' and 'os' properties.
@@ -746,11 +1026,38 @@ If there are no records of a given type in the zone, the call will croak with
 an error message about an invalid method. (This is not an ideal behavior, but
 has been kept for backwards compatibility.)
 
+The 'ORIGIN' property is the fully-qualified origin of the record. See
+L<fqname> for details on constructing a fully qualified domain name. Note: for
+SOA records, the 'ORIGIN' will match the 'origin' property when the SOA record
+is specified as fully qualified.
+
 =item soa()
 
 Returns a hash reference with the following properties:
 'serial', 'origin', 'primary', 'refresh', 'retry', 'ttl', 'minimumTTL',
-'email', 'expire'
+'email', 'expire', 'class', 'ORIGIN'.
+
+The 'ORIGIN' property is returned separate from 'origin' property, though the
+data may be the same. 'ORIGIN' represents the implicit origin for the record
+while 'origin' represents the origin specified on the SOA line in the file.
+
+If the 'origin' value is relative (that is, does not end with a '.'), the
+actual zone for which the SOA line applies must be computed by concatenating
+the 'origin' and 'ORIGIN' values. See L<fqname> for details. If the 'origin'
+value is absolute, no computation is necessary and 'origin' is the same as
+'ORIGIN'.
+
+=item generate()
+
+Returns an array of hashes representing $GENERATE directives present in the
+zone. Note, $GENERATE directives are BIND-specific additions. They are not
+expanded by DNS::ZoneParse, but users are able to access and modify these
+directives. The following properties are returned:
+
+'range', 'lhs', 'ttl', 'class', 'type', 'rhs', 'ORIGIN'.
+
+See the BIND documentation for details on the syntax and usage of the $GENERATE
+directive.
 
 =item dump
 
@@ -758,6 +1065,18 @@ Returns a copy of the datastructute that stores all the resource records. This
 might be useful if you want to quickly transform the data into another format,
 such as XML.
 
+=item fqname
+
+Takes a single parameter, a hash reference containing a record.
+
+Returns the fully qualified name of this record, with a trailing '.'. In most
+cases this is as simple as concatenating the 'name' and 'ORIGIN' with a '.'
+unless 'name' is '@', in which case the fqname is simply the 'ORIGIN'. For
+SOA records, the same process is performed on the 'origin' instead of 'name'.
+
+Please note, fqname will not expand the right hand side of a record (ie,
+CNAME, SOA, MX, etc). The user must expand these values via the above method.
+
 =item new_serial
 
 C<new_serial()> incriments the Zone serial number. It will generate a
@@ -797,19 +1116,19 @@ DNS::ZoneParse will C<croak> when an unparsable line is encountered, but will
 continue to parse the file. Each time an unparsable line is encountered, an
 internal counter is incrememnted. See C<last_parse_error_count> for details.
 
+The callback is passed four parameters, a reference to the DNS::ZoneParse
+object which is doing the parsing, the text of the line that is unable to be
+parsed, the text of the reason the line could not be parsed, and the text of
+the last successfully parsed line.
+
 If you want to abort parsing when an unparsable line is found, call C<die>
 from within your callback and catch that die with an eval block around the
 DNS::ZoneParse constructor (or call to _parse).
 
-Takes a single optional parameter, a code reference to the function that
-will be called when an unparsable line is reached. This code reference will be
-passed two parameters, the first is a reference to the DNS::ZoneParse object
-that is parsing the file, and the second is the text of the line that could
-not be parsed. Returns a reference to the last callback.
-
-If passed an undefined value, a reference to the current callback is returned.
-
-If passed any other value, undef is returned.
+The method takes a single optional parameter, a code reference to the function
+that will be called when an unparsable line is reached. Returns a reference to
+the last callback. If passed an undefined value, a reference to the current
+callback is returned. If passed any other value, undef is returned.
 
 =back
 
